/*
 * Copyright (c) 2001, Swedish Institute of Computer Science.
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution. 
 *
 * 3. Neither the name of the Institute nor the names of its contributors 
 *    may be used to endorse or promote products derived from this software 
 *    without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE 
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE. 
 *
 * This file is part of the lwIP TCP/IP stack.
 * 
 * Author: Adam Dunkels <adam@sics.se>
 *
 * $Id: udp.c,v 1.10 2001/02/06 10:47:32 adam Exp $
 */

/*-----------------------------------------------------------------------------------*/
/* udp.c
 *
 * The code for the User Datagram Protocol UDP.
 *
 */
/*-----------------------------------------------------------------------------------*/
#include "lwip/debug.h"

#include "lwip/def.h"
#include "lwip/mem.h"
#include "lwip/inet.h"
#include "lwip/netif.h"
#include "lwip/udp.h"
#include "lwip/icmp.h"

#include "lwip/stats.h"

/*-----------------------------------------------------------------------------------*/

/* The list of UDP PCBs. */
static struct udp_pcb *udp_pcbs = NULL;
/*-----------------------------------------------------------------------------------*/
void
udp_init(void)
{
}
/*-----------------------------------------------------------------------------------*/
void
udp_input(struct pbuf *p)
{
  struct udp_hdr *udphdr;  
  struct udp_pcb *pcb;
  struct ip_hdr *iphdr;

#ifdef UDP_STATS
  stats.udp.recv++;
#endif /* UDP_STATS */
  
  iphdr = (struct ip_hdr *)((int)p->payload - IP_HLEN);
  udphdr = p->payload;

#ifdef IPv6
  if(iphdr->nexthdr == IP_PROTO_UDPLITE) {    
#else
  if(iphdr->proto == IP_PROTO_UDPLITE) {    
#endif /* IPv4 */
    /* Do the UDP Lite checksum */
    if(inet_chksum_pseudo(p, (struct ip_addr *)&(iphdr->src),
			  (struct ip_addr *)&(iphdr->dest),
			  IP_PROTO_UDPLITE, p->tot_len) != 0) {
      DEBUGF(UDP_DEBUG, ("udp_input: UDP Lite datagram discarded due to failing checksum\n"));
#ifdef UDP_STATS
      stats.udp.chkerr++;
      stats.udp.drop++;
#endif /* UDP_STATS */
      return;
    }      
  } else {
    if(udphdr->chksum != 0) {
      if(inet_chksum_pseudo(p, (struct ip_addr *)&(iphdr->src),
			    (struct ip_addr *)&(iphdr->dest),
			    IP_PROTO_UDP, p->tot_len) != 0) {
	DEBUGF(UDP_DEBUG, ("udp_input: UDP datagram discarded due to failing checksum\n"));

#ifdef UDP_STATS
	stats.udp.chkerr++;
	stats.udp.drop++;
#endif /* UDP_STATS */
	return;
      }
    }
  }

  /*  p->payload = (char *)p->payload + UDP_HLEN;
  p->len -= UDP_HLEN;
  p->tot_len -= UDP_HLEN;*/

  pbuf_header(p, -UDP_HLEN);
  
  udphdr->src = ntohs(udphdr->src);
  udphdr->dest = ntohs(udphdr->dest);
  /*  udphdr->len = ntohs(udphdr->len);*/

#if UDP_DEBUG
  udp_debug_print(udphdr);
#endif /* UDP_DEBUG */

  /* demultiplex packet */

  for(pcb = udp_pcbs; pcb != NULL; pcb = pcb->next) {
    DEBUGF(UDP_DEBUG, ("udp_input: pcb local port %d (dgram %d)\n",
		       pcb->local_port, udphdr->dest));
    if(pcb->local_port == udphdr->dest) {
      pcb->recv(pcb->recv_arg, pcb, p, &(iphdr->src), udphdr->src);
      return;
    }
  }

  /* no match was found, send ICMP destination port unreachable */

  /* deconvert from host to network byte order */
  udphdr->src = htons(udphdr->src);
  udphdr->dest = htons(udphdr->dest); 
  /*  udphdr->len = ntohs(udphdr->len);*/

  /* adjust pbuf pointer */
  p->payload = iphdr;
  icmp_dest_unreach(p, ICMP_DUR_PORT);
#ifdef UDP_STATS
  stats.udp.proterr++;
  stats.udp.drop++;
#endif /* UDP_STATS */  
  pbuf_free(p);
}
/*-----------------------------------------------------------------------------------*/
static void
udp_output(struct pbuf *p, struct ip_addr *src_ip, struct ip_addr *dest_ip,
	   unsigned short src_port, unsigned short dest_port, unsigned int flags)
{
  struct udp_hdr *udphdr;
  struct netif *netif;
      
  if(pbuf_header(p, UDP_HLEN)) {
    DEBUGF(UDP_DEBUG, ("ip_output: not enough room for UDP header in pbuf\n"));
    pbuf_free(p);
#ifdef UDP_STATS
    stats.udp.err++;
#endif /* UDP_STATS */
    return;
  }

  udphdr = p->payload;
  udphdr->src = htons(src_port);
  udphdr->dest = htons(dest_port);
  udphdr->len = htons(p->len + UDP_HLEN);
  udphdr->chksum = 0;


  if(!ip_addr_isany(src_ip)) {
    if((netif = ip_route(dest_ip)) == NULL) {
      DEBUGF(UDP_DEBUG, ("udp_send: No route to 0x%lx\n", dest_ip->addr));
      pbuf_free(p);
#ifdef UDP_STATS
      stats.udp.rterr++;
#endif /* UDP_STATS */
      return;
    }
    src_ip = &(netif->ip_addr);
  }
  
  /* calculate checksum */
  if((flags & UDP_FLAGS_NOCHKSUM) == 0) {

    udphdr->chksum = inet_chksum_pseudo(p,
					src_ip, dest_ip,
					IP_PROTO_UDP, p->tot_len);    
  }

#ifdef UDP_STATS
  stats.udp.xmit++;
#endif /* UDP_STATS */
  
  ip_output_if(p, src_ip, dest_ip, UDP_TTL, IP_PROTO_UDP, netif);
  pbuf_free(p);
}
/*-----------------------------------------------------------------------------------*/
void
udp_send(struct udp_pcb *pcb, struct pbuf *p)
{
  udp_output(p, IP_ADDR_ANY, &(pcb->dest_ip), pcb->local_port, pcb->dest_port, pcb->flags);
}
/*-----------------------------------------------------------------------------------*/
void
udp_bind(struct udp_pcb *pcb, struct ip_addr *ipaddr, unsigned short port)
{
  struct udp_pcb *ipcb;
  if(ipaddr != NULL) {
    pcb->local_ip = *ipaddr;
  }
  pcb->local_port = port;
  
  for(ipcb = udp_pcbs; ipcb != NULL; ipcb = ipcb->next) {
    if(pcb == ipcb) {
      return;
    }
  }

  pcb->next = udp_pcbs;
  udp_pcbs = pcb;
}
/*-----------------------------------------------------------------------------------*/
void
udp_connect(struct udp_pcb *pcb, struct ip_addr *ipaddr, unsigned short port)
{
  struct udp_pcb *ipcb;
  if(ipaddr != NULL) {
    pcb->dest_ip = *ipaddr;
  }
  pcb->dest_port = port;
  
  for(ipcb = udp_pcbs; ipcb != NULL; ipcb = ipcb->next) {
    if(pcb == ipcb) {
      return;
    }
  }
  pcb->next = udp_pcbs;
  udp_pcbs = pcb;
}
/*-----------------------------------------------------------------------------------*/
void
udp_recv(struct udp_pcb *pcb,
	 void (* recv)(void *arg, struct udp_pcb *pcb, struct pbuf *p,
		       struct ip_addr *addr, u16_t port),
	 void *recv_arg)
{
  pcb->recv = recv;
  pcb->recv_arg = recv_arg;
}
/*-----------------------------------------------------------------------------------*/
void
udp_pcb_remove(struct udp_pcb *pcb)
{
  struct udp_pcb *pcb2;
  
  if(udp_pcbs == pcb) {
    udp_pcbs = udp_pcbs->next;
  } else for(pcb2 = udp_pcbs; pcb2 != NULL; pcb2 = pcb2->next) {
    if(pcb2->next != NULL && pcb2->next == pcb) {
      pcb2->next = pcb->next;
    }
  }

  mem_free(pcb);  
}
/*-----------------------------------------------------------------------------------*/
struct udp_pcb *
udp_pcb_new(void) {
  struct udp_pcb *pcb;
  pcb = mem_malloc(sizeof(struct udp_pcb));
  if(pcb != NULL) {
    bzero(pcb, sizeof(struct udp_pcb));
    return pcb;
  }
  return NULL;

}
/*-----------------------------------------------------------------------------------*/
#if UDP_DEBUG
void
udp_debug_print(struct udp_hdr *udphdr)
{
  DEBUGF(UDP_DEBUG, ("UDP header:\n"));
  DEBUGF(UDP_DEBUG, ("+-------------------------------+\n"));
  DEBUGF(UDP_DEBUG, ("|     %5d     |     %5d     | (src port, dest port)\n",
		     ntohs(udphdr->src), ntohs(udphdr->dest)));
  DEBUGF(UDP_DEBUG, ("+-------------------------------+\n"));
  DEBUGF(UDP_DEBUG, ("|     %5d     |     0x%04x    | (len, chksum)\n",
		     ntohs(udphdr->len), ntohs(udphdr->chksum)));
  DEBUGF(UDP_DEBUG, ("+-------------------------------+\n"));
}
#endif /* UDP_DEBUG */
/*-----------------------------------------------------------------------------------*/










